#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

typedef struct {
unsigned char *code;
unsigned long size;
enum {TI92PLUS=1, TI89=3, V200=8} output;
} obj;

// These are the hunk IDs from the A68K source
#define HunkUnit 999
#define HunkName 1000
#define HunkCode 1001
#define HunkData 1002
#define HunkBSS  1003
#define HunkR32  1004
#define HunkR16  1005
#define HunkR8   1006
#define HunkExt  1007
#define HunkSym  1008
#define HunkDbg  1009
#define HunkEnd  1010
// Symbol entry type IDs
#define TypeEnd 0
#define TypeExport 1
#define TypeImport32 0x81
#define TypeImport16 0x83
#define TypeImport8 0x84
#define TypeImport16Alt 0x8a
#define TypeEqu 0x02

#define EXTRA_INFO(x)
#define ERROR(x) {printf(x); exit(1); }
#define ERROR1(x,y) {printf(x,y); exit(1); }
#define ERROR2(x,y,z) {printf(x,y,z); exit(1); }

void removeExt(const char *source, char *dest)
{ int i = 0; while (source[i]!=0 & source[i]!='.') dest[i++] = source[i]; dest[i] = 0; }

long ReadDWord(FILE *fp)
{ long v=fgetc(fp)<<24; v|=fgetc(fp)<<16; v|=fgetc(fp)<<8; v|=fgetc(fp); return v; }

void WriteDWord(FILE *fp, long x)
{ fputc(x>>24, fp); fputc((x>>16)&0xFF, fp); fputc((x>>8)&0xFF, fp); fputc(x&0xFF, fp);}

void WriteDWordRAM(unsigned char *a, unsigned long x)
{ *a++ = x>>24; *a++ = (x>>16)&0xFF; *a++ = (x>>8)&0xFF; *a++ = x&0xFF;}

void WriteWord(FILE *fp, long x)
{ fputc((x>>8)&0xFF, fp); fputc(x&0xFF, fp);}

void WriteByte(FILE *fp, long x)
{ fputc(x, fp); }

void WriteLong(FILE *fp, long x)
{ fputc(x>>24, fp); fputc((x>>16)&0xFF, fp); fputc((x>>8)&0xFF, fp); fputc(x&0xFF, fp);}

obj	ReadAmigaObj(const char *filename)
{
   obj o = {NULL, 0, 0};
   unsigned int size, type, i, ofs;
   char sym[100];
   FILE *fp = fopen(filename, "rb");
   if (fp == NULL) return o;
   
   // Process each hunk in the object file
   while (!feof(fp))
   	{
	int hunk=ReadDWord(fp);
	if (hunk==HunkCode)		// Code
		{
		if (o.code) ERROR("Duplicate CODE section");
		o.size = ReadDWord(fp) * 4; // Size is in DWORDs
		o.code = malloc(o.size);
		if (o.code == NULL) ERROR("Memory overflow");
		fread(o.code, o.size, 1, fp);
		}
	else if (hunk==HunkExt) 	// Exports references
		while (1)
			{
			size=ReadDWord(fp);
			// High byte of size contains type byte
			type=(size>>24)&0xff;
			size&=0xffffff;
			if (type==TypeEnd)
				break;
			// Read name of symbol
			for (i=0;i<(size<<2);i++)
				sym[i]=fgetc(fp);
			sym[i]=0;
			if (type==TypeExport)
				{
				// Get offset of exported function
				ofs=ReadDWord(fp);
				// Check for special export names
				if (!strcmp(sym,"_ti89"))
					o.output = TI89;
				else if (!strcmp(sym,"_ti92plus"))
					o.output = TI92PLUS;
				else if (!strcmp(sym,"_v200"))
					o.output = V200;
				}
			else	ERROR2("Unsupported Type: %d for symbol: %s", type, sym);
			}
	else if (hunk==HunkR32)
		{
		// Read each section's relocations
		while (1)
			{
			int count=ReadDWord(fp);
			if (!count) break;
			ReadDWord(fp);
            for(i = 0 ; i < count; i++)
					ReadDWord(fp);
			}
		}
	else if (hunk==HunkR16 || hunk==HunkR8)
		{ERROR("Error: 16-bit & 8-bit relocations are unsupported\n");}
	else if (hunk==HunkEnd) // End of code or BSS section
		{
			break;
		}
	else 	{
		size=ReadDWord(fp); // Size is in DWORDs
		fseek(fp,size*4,SEEK_CUR);
		}
	}

   fclose(fp);
   return o;
}

char	TIBProductName[] = "Advanced Mathematics Software";
char	TIBSignature1[64];

int	WriteTIB2(obj o, const char *filename)
{
	FILE	*f;
	if (o.code == NULL)	return -1;
	if (o.size&1)		return -2;
	f = fopen(filename, "wb");
	if (f == NULL)		return -3;
	
	WriteWord(f, 0x800F); WriteLong(f, o.size + 126+8);
	WriteWord(f, 0x8011); WriteByte(f, o.output);
	WriteWord(f, 0x8021); WriteByte(f, 0x03);
	WriteWord(f, 0x8032); WriteWord(f, 0x384F);
	WriteWord(f, 0x80A1); WriteByte(f, 0x01);
	WriteWord(f, 0x804D); WriteByte(f, sizeof(TIBProductName)-1); fwrite(TIBProductName, sizeof(TIBProductName)-1, 1, f);
	WriteWord(f, 0x0326); WriteLong(f, 0x01020304); WriteWord(f, 0x0506);
	WriteWord(f, 0x020D); WriteByte(f, 0x40); fwrite(TIBSignature1, 64, 1, f);
	WriteWord(f, 0x807F); WriteLong(f, o.size + 8); 
		WriteLong(f, 0xCCCCCCCC);
		fwrite(o.code, o.size, 1, f);
		WriteLong(f, 0x00000000);
	WriteWord(f, 0x020D); WriteByte(f, 0x40); fwrite(TIBSignature1, 64, 1, f);
	fclose(f);
}

int	main(int argc, char *argv[])
{
	obj o;
	char	Name[100];
	
	if (argc != 2)
		ERROR("ERROR: maketib obj.o\n")
	o = ReadAmigaObj(argv[1]);
	if (o.code == NULL)
		ERROR1("ERROR: Bad object file %s\n", argv[1]);
	removeExt(argv[1], Name);
	strcat(Name, ".tib");
	WriteTIB2(o, Name);
	exit(0);
}
